/*
FILE:     TableFile.h
*/
/*
VERSION:  2.007
*/
/*
DATE:     4/3/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


#ifndef TABLEFILE_H
#define TABLEFILE_H


#include <vector>
#include <set>

#ifndef VLAD_DECIDE
#include "mapped_vector.h"
#include "mapped_vector.C"
#endif
#include "mapped_ptr_vector.h"
#include "mapped_ptr_vector.C"
#include "ISTable.h"
#include "FileNavigator.h"
#include "FileNavigatorError.h"


/**
**  Block class represents data blocks of CIF files.
*/
class Block
{
  private:
    string _name;
    eFileMode _fileMode;
    FileNavigator* _fileNav;

    ISTable* _GetTablePtr(const unsigned int tableIndex);

  public:
    mapped_ptr_vector<ISTable, StringCompare> _tables;

    Block(const string& name, FileNavigator* fileNav,
      const eFileMode fileMode = READ_MODE, const StringCompare::eCompareType
      caseSense = StringCompare::eCASE_SENSITIVE);
    Block(const Block& t);
    ~Block();

    vector<pair<string, ISTable::eTableDiff> > operator==(Block& inBlock);

    inline void SetName(const string& name);
    inline string GetName() const;

    /// Adds a new table, along with its file index and its table pointer
    void AddTable(const string& name,
      const int indexInFile = 0, ISTable* isTableP = NULL);

    void GetTableNames(vector<string>& names);
    bool IsTablePresent(const string& name);

    ISTable* GetTablePtr(const string& name);

    void DeleteTable(const string& name);

    /// Updates the existing table or adds a new one
    void WriteTable(ISTable* isTableP);

    void Print();
};


/**
**  TableFile class represents a file composed of data blocks.
*/
class TableFile
{
  private:
    static const string _version;
    void Init();
    int Open(const string& fileName, const eFileMode fileMode);

  protected:
    string _fileName;

    eFileMode _fileMode;

    // Indicates case sensitivity of identifiers
    StringCompare::eCompareType _caseSense;

    // Indicates the current status of the object
    unsigned int _statusInd;  

    mapped_ptr_vector<Block, StringCompare> _blocks;

    FileNavigator* _f;

    void _SetStatusInd(const string& blockName);

    void _AddBlock(const string& blockName, FileNavigator* fileNav);

    void _GetNumTablesInBlocks(vector<Word>& numTablesInBlocks);

    ISTable* _GetTablePtr(const unsigned int blockIndex,
      const unsigned int tableIndex);
    void _GetAllTables();

    unsigned int GetTotalNumTables();
    void GetTableNames(vector<string>& tableNames);

    void GetTablesIndices(vector<unsigned int>& tablesIndices);
    void GetSortedTablesIndices(vector<unsigned int>& tablesIndices);

    int _ReadFileIndex();
    int _ReadFileIndexVersion0();
    int _ReadFileIndexVersion1();
    int _WriteFileIndex(FileNavigator* fileNav, const vector<int>& tableLocs);

  public:
    enum eStatusInd
    {
        eCLEAR_STATUS = 0x0000,
        eDUPLICATE_BLOCKS = 0x0001,
        eUNNAMED_BLOCKS = 0x0002
    };

    TableFile(const StringCompare::eCompareType caseSense =
      StringCompare::eCASE_SENSITIVE);

    TableFile(const eFileMode fileMode, const string& fileName,
      const StringCompare::eCompareType caseSense =
      StringCompare::eCASE_SENSITIVE);

    virtual ~TableFile();

    inline string GetFileName(void);
    inline eFileMode GetFileMode(void);
    inline StringCompare::eCompareType GetCaseSensitivity(void);
    inline unsigned int GetStatusInd(void);

    /// Adds a new block. Returns the name of a newly created block.
    string AddBlock(const string& blockName);

    bool IsBlockPresent(const string& blockName);

    inline unsigned int GetNumBlocks();
    string GetFirstBlockName();
    void GetBlockNames(vector<string>& blockNames);
#ifndef VLAD_DECIDE
    void GetBlockNames(mapped_vector<string>& blockNames);
#endif
    void RenameBlock(const string& oldBlockName, const string& newBlockName);
    inline void RenameFirstBlock(const string& newBlockName);

    Block& GetBlock(const string& blockName);
 
    void Flush();

    void Serialize(const string& fileName);

    void Close();

    void PrintHeaderInfo();

    inline void DumpFile(void);
    inline void PrintIndex(void);
    inline void PrintIndexPosition(int position);
};


inline void Block::SetName(const string& name)
{
    _name = name;
}


inline string Block::GetName() const
{
    return _name;
} 


inline string TableFile::GetFileName(void)
{
    return _fileName;
}


inline eFileMode TableFile::GetFileMode(void)
{
    return _fileMode;
}


inline StringCompare::eCompareType TableFile::GetCaseSensitivity(void)
{
    return(_caseSense);
}


inline unsigned int TableFile::GetStatusInd(void)
{
    return _statusInd;
}

   
inline unsigned int TableFile::GetNumBlocks()
{
    return _blocks.size();
}


inline void TableFile::RenameFirstBlock(const string& newBlockName)
{
    RenameBlock(GetFirstBlockName(), newBlockName);
}


inline void TableFile::DumpFile(void)
{
    if (_f)
        _f->DumpFile();
}


inline void TableFile::PrintIndex(void)
{
    if (_f)
        _f->PrintIndex();
}


inline void TableFile::PrintIndexPosition(int position)
{
    if (_f)
        _f->PrintIndexPosition(position);
}


#endif // TABLEFILE_H
