/*
FILE:     STable.C
*/
/*
VERSION:  2.007
*/
/*
DATE:     4/3/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


#include <stdexcept>
#include <string>
#include <iostream>
#include <iomanip>

#include "Exceptions.h"
#include "GenString.h"
#include "STable.h"


using namespace std;


STable::STable(const StringCompare::eCompareType colCaseSense) :
  _colCaseSense(colCaseSense), _numRows(0), _numNonEmptyCols(0)
{


}


STable::STable(const string& name,
  const StringCompare::eCompareType colCaseSense) : _name(name),
  _colCaseSense(colCaseSense), _numRows(0), _numNonEmptyCols(0)
{


}


STable::STable(const STable& inTable)
{

    _name = inTable._name;
    _colCaseSense = inTable._colCaseSense;
    _colNames = inTable._colNames;
    _numRows = inTable._numRows;
    _numNonEmptyCols = inTable._numNonEmptyCols;

    // This statement only copies pointer addresses and not the column
    // vectors referenced by those pointers.
    _data = inTable._data;

    // Here, column vectors themselves are copied and previously set
    // pointers are properly initialized.

    for (unsigned int colIndex = 0; colIndex < _colNames.size(); ++colIndex)
    {
        _data[colIndex] = new vector<string>(*(inTable._data[colIndex]));
    }

}


STable::~STable()
{

    Clear();

}


STable& STable::operator=(const STable& inTable)
{

    if (this != &inTable)
    {
        Clear();

        _name = inTable._name;
        _colCaseSense = inTable._colCaseSense;
        _colNames = inTable._colNames;
        _numRows = inTable._numRows;
        _numNonEmptyCols = inTable._numNonEmptyCols;

        // This statement only copies pointer addresses and not the column
        // vectors referenced by those pointers.
        _data = inTable._data;

        // Here, column vectors themselves are copied and previously set
        // pointers are properly initialized.

        for (unsigned int colIndex = 0; colIndex < _colNames.size();
          ++colIndex)
        {
            _data[colIndex] = new vector<string>(*(inTable._data[colIndex]));
        }
    }

    return(*this);

}


void STable::Clear()
{

    if (!_data.empty())
    {
        for (unsigned int colIndex = 0; colIndex < _data.size(); colIndex++)
        {
            _data[colIndex]->clear();
            delete _data[colIndex];
        }

        _data.clear();
    }

    _numRows = 0;

    _numNonEmptyCols = 0; 

    _colNames.clear();

}


void STable::Rename(const string& name)
{

    _name = name;

}


void STable::GetColumnNames(vector<string>& colNames) const
{

    colNames = _colNames;

}


bool STable::IsColumnPresent(const string& colName)
{

    try
    {
        GetColumnIndex(colName);

        return(true);
    }
    catch (NotFoundException)
    {
        return(false);
    }

}


void STable::AddColumn(const string& colName, const vector<string>& col)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::AddColumn");
    }

    if (col.size() > _numRows)
    {
        throw out_of_range("Column size greater than number of rows in "\
          " STable::AddColumn");
    }

    STable::InsertColumn(colName, _colNames.size(), col);

}


void STable::InsertColumn(const string& colName, const string& atColName,
  const vector<string>& col)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::InsertColumn");
    }

    if (col.size() > _numRows)
    {
        throw out_of_range("Column size greater than number of rows in "\
          " STable::InsertColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(atColName);

        STable::InsertColumn(colName, colIndex, col);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::InsertColumn");
        throw;
    }

}


void STable::FillColumn(const string& colName, const vector<string>& col)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::FillColumn");
    }

    if (col.empty())
    {
        return;
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        STable::FillColumn(col, colIndex);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::FillColumn");
        throw;
    }

}


void STable::AppendToColumn(const string& colName, const string& cell)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::AppendToColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        STable::AppendToColumn(cell, colIndex);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::AppendToColumn");
        throw;
    }

}


void STable::AppendToColumn(const string& colName, const vector<string>& col)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::AppendToColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        STable::AppendToColumn(col, colIndex);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::AppendToColumn");
        throw;
    }

}


void STable::GetColumn(vector<string>& col, const string& colName)
{

    col.clear();

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::GetColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        col = *_data[colIndex];
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::GetColumn");
        throw;
    }

}


void STable::GetColumn(vector<string>& subCol,
  const string& colName, const unsigned int fromRowIndex,
  unsigned int toRowIndex)
{

    subCol.clear();

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::GetColumn");
    }

    if (toRowIndex < fromRowIndex)
    {
        // To index less than from index.
        throw out_of_range("From row index greater than to row index "\
          "STable::GetColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        if (colIndex >= _numNonEmptyCols)
        {
            // Empty column.
            throw EmptyValueException("Empty column",
              "STable::GetColumn");
        }

        if (fromRowIndex >= _numRows)
        {
            // Wrong from row index.
            throw out_of_range("Invalid from row index in "\
              "STable::GetColumn");
        }

        if (toRowIndex >= _numRows)
        {
            // Wrong to row index.
            throw out_of_range("Invalid to row index in "\
              "STable::GetColumn");
        }

        // toRowIndex is inclusive, so one more spot for it needs
        // to be reserved
        subCol.reserve(toRowIndex - fromRowIndex + 1);

        copy((*_data[colIndex]).begin() + fromRowIndex,
          (*_data[colIndex]).begin() + toRowIndex + 1,
          back_inserter(subCol));
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::GetColumn");
        throw;
    }

}


void STable::GetColumn(vector<string>& subCol,
  const string& colName, const vector<unsigned int>& rowIndices)
{

    subCol.clear();

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::GetColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        if (colIndex >= _numNonEmptyCols)
        {
            // Empty column.
            throw EmptyValueException("Empty column",
              "STable::GetColumn");
        }

        for (unsigned int index = 0; index <  rowIndices.size(); ++index)
        {
            if (rowIndices[index] >= _numRows)
            {
                // Wrong row index.
                throw out_of_range("Invalid row index in "\
                  "STable::GetColumn");
            }
        }

        subCol.reserve(rowIndices.size());

        for (unsigned int index = 0; index <  rowIndices.size(); index++)
        {
            subCol.push_back((*_data[colIndex])[rowIndices[index]]);
        }
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::GetColumn");
        throw;
    }

}


unsigned int STable::ColumnLength(const string& colName) const
{

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        if (colIndex < _numNonEmptyCols)
        {
            return(_numRows);
        }
        else
        {
            return(0);
        }
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::ColumnLength");
        throw;
    }

}


void STable::RenameColumn(const string& oldColName, const string& newColName)
{

    if (oldColName.empty())
    {
        // Empty old column name.
        throw EmptyValueException("Empty old column name",
          "STable::RenameColumn");
    }

    if (newColName.empty())
    {
        // Empty new column name.
        throw EmptyValueException("Empty new column name",
          "STable::RenameColumn");
    }

    if (IsColumnPresent(newColName))
    {
        // Column already exists
        throw AlreadyExistsException("Duplicate column name",
          "STable::RenameColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(oldColName);

        _colNames[colIndex] = newColName;
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Old column not found",
          "STable::RenameColumn");
        throw;
    }

}


void STable::ClearColumn(const string& colName)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::ClearColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        STable::ClearColumn(colIndex);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::ClearColumn");
        throw;
    }

}


void STable::DeleteColumn(const string& colName)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::DeleteColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        STable::DeleteColumn(colIndex);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::DeleteColumn");
        throw;
    }

}


unsigned int STable::GetColumnIndex(const string& colName) const
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::GetColumnIndex");
    }

    unsigned int colIndex;

    for (colIndex = 0; colIndex < _colNames.size(); ++colIndex)
    {
        if (String::IsEqual(_colNames[colIndex], colName, _colCaseSense))
        {
            break;
        }
    }

    if (colIndex == _colNames.size())
    {
        // Column not found.
        throw NotFoundException("Column not found",
          "STable::GetColumnIndex");
    }
    else
    {
        return(colIndex);
    }

}


void STable::GetColumnsIndices(vector<unsigned int>& colIndices,
  const vector<string>& colNames)
{

    if (colNames.empty())
    {
        _GetColumnsIndices(colIndices, _colNames);
    }
    else
    {
        _GetColumnsIndices(colIndices, colNames);
    }

}


unsigned int STable::AddRow(const vector<string>& row)
{

    return(STable::InsertRow(_numRows, row));

}


unsigned int STable::InsertRow(const unsigned int atRowIndex,
  const vector<string>& row)
{

    if (_colNames.empty())
    {
        // Wrong row index.
        throw EmptyContainerException("Table has no columns",
          "STable::InsertRow");
    }

    if (atRowIndex > _numRows)
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::InsertRow");
    }

    if (row.size() > _numNonEmptyCols)
    {
        throw out_of_range("Invalid row size in STable::InsertRow");
    }

    unsigned int numCells;

    if (_numRows == 0)
    {
        numCells = _colNames.size();
        _numNonEmptyCols = numCells;
    }
    else
    {
        numCells = _numNonEmptyCols;
    }

    for (unsigned int colIndex = 0; colIndex < numCells; ++colIndex)
    {
        _data[colIndex]->insert(_data[colIndex]->begin() + atRowIndex, 1,
          String::Empty);
    }

    _numRows++;

    STable::FillRow(atRowIndex, row);

    return(_numRows);

}


void STable::FillRow(const unsigned int rowIndex, const vector<string>& row)
{

    if (rowIndex >= _numRows)
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::FillRow");
    }

    if (row.empty())
    {
        return;
    }

    if (row.size() > _numNonEmptyCols)
    {
        throw out_of_range("Invalid row size in STable::FillRow");
    }

    for (unsigned int colIndex = 0; colIndex < _numNonEmptyCols; ++colIndex)
    {
        (*_data[colIndex])[rowIndex] = row[colIndex];
    }

}


void STable::GetRow(vector<string>& row, const unsigned int rowIndex,
  const string& fromColName, const string& toColName)
{

    row.clear();

    try
    {
        unsigned int fromColIndex = 0;
        if (!fromColName.empty())
        {
            fromColIndex = GetColumnIndex(fromColName);
        }

        unsigned int toColIndex = _numNonEmptyCols - 1;
        if (!toColName.empty())
        {
            toColIndex = GetColumnIndex(toColName);
        }

        STable::GetSubRow(row, rowIndex, fromColIndex, toColIndex + 1);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::GetRow");
        throw;
    }

}


void STable::ClearRow(const unsigned int rowIndex)
{

    if (rowIndex >= _numRows)
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::ClearRow");
    }

    for (unsigned int colIndex = 0; colIndex < _numNonEmptyCols; ++colIndex)
    {
        (*_data[colIndex])[rowIndex].clear();
    }

}


void STable::DeleteRow(const unsigned int rowIndex)
{

    if (rowIndex >= _numRows)
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::DeleteRow");
    }

    for (unsigned int colIndex = 0; colIndex < _numNonEmptyCols; ++colIndex)
    {
        _data[colIndex]->erase(_data[colIndex]->begin() + rowIndex);
    }

    _numRows--;

    if (_numRows == 0)
    {
        _numNonEmptyCols = 0;
    }

}


void STable::UpdateCell(const unsigned int rowIndex, const string& colName,
  const string& cell)
{

    if (rowIndex >= _numRows)
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::UpdateCell");
    }

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::UpdateCell");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        if (colIndex >= _numNonEmptyCols)
        {
            // Wrong column index.
            throw out_of_range("Invalid column index in STable::UpdateCell");
        }

        STable::UpdateCell(cell, colIndex, rowIndex);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::UpdateCell");
        throw;
    }

}


const string& STable::operator()(const unsigned int rowIndex,
  const string& colName) const
{

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        if (rowIndex >= _numRows)
        {
            // Wrong row index.
            throw out_of_range("Invalid row index in STable::operator()");
        }

        if (colIndex >= _numNonEmptyCols)
        {
            // Wrong column index.
            throw out_of_range("Invalid column index in STable::operator()");
        }

        return((*_data[colIndex])[rowIndex]);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "STable::operator()");
        throw;
    }

}


ostream& operator<<(ostream& out, const STable& sTable)
{

    vector<string> colNames;

    sTable.GetColumnNames(colNames);

    unsigned int numRows = sTable.GetNumRows();

    out << endl << "Table " << sTable.GetName() <<  "  Columns: " <<
      colNames.size() << "  Rows: " << numRows << endl << endl;

    for (unsigned int i = 0; i < colNames.size(); i++)
    {
        out << setw(10) << colNames[i] << " ";
    }
    out << endl;

    for (unsigned int j = 0; j < numRows; j++)
    {
        for (unsigned int i = 0; i < colNames.size(); i++)
        {
            if (sTable.ColumnLength(colNames[i]) > j)
            {
                out << setw(10) << sTable(j, colNames[i]) << " ";
            }
            else
            {
                out << setw(10) << " " << " ";
            }
        }

        out << endl;
    }

    return(out);

}


void STable::InsertColumn(const string& colName, const unsigned int colIndex,
  const vector<string>& col)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::InsertColumn");
    }

    if (colIndex > _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::InsertColumn");
    }

    if ((_numRows != 0) && (col.size() > _numRows))
    {
        // Wrong column index.
        throw out_of_range("Invalid column size in STable::FillColumn");
    }

    try
    {
        GetColumnIndex(colName);

        // Column already exists
        throw AlreadyExistsException("Duplicate column name",
          "STable::InsertColumn");
    }
    catch (NotFoundException)
    {
        // Continue execution, as it is normal that column is not found
        _colNames.insert(_colNames.begin() + colIndex, colName);

        _data.insert(_data.begin() + colIndex, new vector<string>);

        STable::CreateColumn(colIndex, col);
    }

}


void STable::FillColumn(const vector<string>& col, const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::FillColumn");
    }

    if (col.empty())
    {
        return;
    }

    if ((_numRows != 0) && (col.size() > _numRows))
    {
        // Wrong column index.
        throw out_of_range("Invalid column size in STable::FillColumn");
    }

    if (_data[colIndex]->empty())
    {
        STable::CreateColumn(colIndex, col);
        return;
    }

    copy(col.begin(), col.end(), _data[colIndex]->begin());

}


void STable::CreateColumn(const unsigned int colIndex,
  const vector<string>& col)
{

    if (colIndex > _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::CreateColumn");
    }

    if ((_numRows != 0) && (col.size() > _numRows))
    {
        // Wrong column size.
        throw out_of_range("Invalid column size in STable::CreateColumn");
    }

    if ((!col.empty()) && (colIndex != 0) && (_data[colIndex - 1]->empty()))
    {
        // Previous column empty. Cannot insert, since rectangularity will be
        // violated.
        throw out_of_range("Previous column empty STable::CreateColumn");
    }

    if (_numRows == 0)
    {
        // For empty table, new column defines the initial number of rows.
        _numRows = col.size();
    }

    if (_data[colIndex]->empty())
    {
        // If table column empty, insert empty cells.
        _data[colIndex]->insert(_data[colIndex]->begin(), _numRows,
          String::Empty);

        if (_numRows != 0)
        {
            _numNonEmptyCols++;
        }
    }

    copy(col.begin(), col.end(), _data[colIndex]->begin());

}


void STable::AppendToColumn(const string& cell, const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::AppendToColumn");
    }

    if ((colIndex != 0) && (_data[colIndex - 1]->empty()))
    {
        // Previous column empty. Cannot append, since rectangularity will be
        // violated.
        throw out_of_range("Previous column empty STable::AppendToColumn");
    }

    STable::AddRow();

    (*_data[colIndex])[_numRows - 1] = cell;

}


void STable::AppendToColumn(const vector<string>& col,
  const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::AppendToColumn");
    }

    if ((colIndex != 0) && (_data[colIndex - 1]->empty()))
    {
        // Previous column empty. Cannot append, since rectangularity will be
        // violated.
        throw out_of_range("Previous column empty STable::AppendToColumn");
    }

    if (col.empty())
    {
        return;
    }

    unsigned int oldNumRows = _numRows;
    for (unsigned int rowI = 0; rowI < col.size(); ++rowI)
    {
        STable::AddRow();
    }

    copy(col.begin(), col.end(), _data[colIndex]->begin() + oldNumRows);

}


void STable::ClearColumn(const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::ClearColumn");
    }

    for (unsigned int rowIndex = 0; rowIndex < _data[colIndex]->size();
      ++rowIndex)
    {
        (*_data[colIndex])[rowIndex].clear();
    }

}


void STable::DeleteColumn(const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::DeleteColumn");
    }

    _data[colIndex]->clear();

    delete _data[colIndex];

    _data.erase(_data.begin() + colIndex);

    _colNames.erase(_colNames.begin() + colIndex);

    if ((colIndex < _numNonEmptyCols) && (_numRows != 0))
    {
        _numNonEmptyCols--;
    }

    // Determines the maximum size of all columns and sets _numRows
    for (unsigned int colIndex = 0; colIndex < _colNames.size(); ++colIndex)
    {
        if (_data[colIndex]->size() > _numRows)
        {
            _numRows = _data[colIndex]->size();
        }
    }

}


void STable::VerifyColumnsIndices(const vector<unsigned int>& colIndices)
{

    for (unsigned int index = 0; index < colIndices.size(); index++)
    {
        if (colIndices[index] >= _colNames.size())
        {
            // Wrong column index.
            throw out_of_range("Invalid column index in "\
              " STable::VerifyColumnsIndices");
        }
    }

}


void STable::_GetColumnsIndices(vector<unsigned int>& colIndices,
  const vector<string>& colNames)
{

    colIndices.clear();

    try
    {
        for (unsigned int index = 0; index < colNames.size(); index++)
        {
            colIndices.push_back(GetColumnIndex(colNames[index]));
        }
    }
    catch (RcsbException& rcsbException)
    {
        colIndices.clear();
        throw;
    }

}


void STable::GetSubRow(vector<string>& subRow, const unsigned int rowIndex,
  const unsigned int fromColIndex, unsigned int toColIndex)
{

    subRow.clear();

    if (rowIndex >= _numRows)
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::GetSubRow");
    }

    if (fromColIndex >= toColIndex)
    {
        // To index less than from index.
        throw out_of_range("Invalid column index range in STable::GetSubRow");
    }

    if (fromColIndex >= _numNonEmptyCols)
    {
        // Wrong from column index.
        throw out_of_range("Invalid from column index in STable::GetSubRow");
    }

    if (toColIndex > _numNonEmptyCols)
    {
        // Wrong to column index.
        throw out_of_range("Invalid to column index in STable::GetSubRow");
    }

    // toColIndex is inclusive, so one more spot for it needs to be reserved
    subRow.reserve(toColIndex - fromColIndex);

    for (unsigned int colIndex = fromColIndex; colIndex < toColIndex;
      colIndex++)
    {
        subRow.push_back((*_data[colIndex])[rowIndex]);
    }

}


void STable::UpdateCell(const string& cell, const unsigned int colIndex,
  const unsigned int rowIndex)
{

    if (colIndex >= _numNonEmptyCols)
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::UpdateCell");
    }

    if (rowIndex >= _numRows)
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::UpdateCell");
    }

    (*_data[colIndex])[rowIndex] = cell;

}


void STable::GetCell(string& cell, const unsigned int colIndex,
  const unsigned int rowIndex)
{

    if (colIndex >= _numNonEmptyCols)
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::GetCell");
    }

    if (rowIndex >= _numRows)
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in STable::GetCell");
    }

    cell = (*_data[colIndex])[rowIndex];

}

