/*
FILE:     CifLoadReorganizer.C
*/
/*
VERSION:  2.007
*/
/*
DATE:     4/3/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


#include <string>

#include "ISTable.h"
#include "DDLParserBase.h"
#include "DICParserBase.h"
#include "CifReorganizerBase.h"
#include "CifLoadReorganizer.h"


string OdbFileVersion("V1");


DdlLoadCont::DdlLoadCont(FileNavigator* fnav, DicFile* dicFile,
  const string& dataBlockName, const string& id, DdlContInfo& inDdlContInfo) :
  DdlCont(dataBlockName, id, inDdlContInfo, fnav), _dic(dicFile)
{
    _verbose = 0;
}


DdlLoadCont::~DdlLoadCont()
{

}


void DdlLoadCont::SetVerbose(int verbose)
{
    _verbose = verbose;
}


int DdlLoadCont::Load()
{
  if (_datablock.empty() || _id.empty())
      return KEYS_NOT_SET; 

  for (unsigned int i = 0; i < contInfo._ddlCats.size(); ++i)
  {
      LoadDdlItems(_ddlItemsStore[i], i);
  }

  return NO_ERROR;
}


void DdlLoadCont::LoadDdlItems(vector<vector<string> >& combo,
  const unsigned int configIndex)
{
    DdlLoadCont::LoadDdlItems(combo, configIndex, _id);    
}


void DdlLoadCont::LoadDdlItems(vector<vector<string> >& combo,
  const unsigned int configIndex, const string& value)
{
    int i;
    ISTable * tbl;
    vector<string> ColumnNames;
    vector<string> values;
    vector<unsigned int> listOut;
    string cell;
    vector<string> columnNames;

    if (_id.empty() || _datablock.empty())
    {
        return;
    }
  
    Block& block = _dic->GetBlock(_datablock);

    tbl = block.GetTablePtr(contInfo._ddlCats[configIndex].ddlCatName);
    if (tbl == NULL)
    {
        return;
    }

    ColumnNames.push_back(contInfo._ddlCats[configIndex].col1);
    if (contInfo._ddlCats[configIndex].nonDefaultValue)
    {
        unsigned int valIndex =
          contInfo.GetDdlItemIndex(DdlContInfo::CIF_DDL_CATEGORY_ITEM_TYPE,
          DdlContInfo::CIF_DDL_ITEM_CODE);

        vector<string> tmp1;
        vector<vector<string> > valueVector;

        for (unsigned int j = 0; j < contInfo._ddlCats[valIndex].\
          ddlItems.size(); ++j)
        {
            valueVector.push_back(tmp1);
        }

        LoadDdlItems(valueVector, valIndex);

        if (valueVector[0].empty())
            return;
        values.push_back(valueVector[0][0]); 

        if (_ddlItemsStore[valIndex][0].empty())
            _ddlItemsStore[valIndex][0].push_back(valueVector[0][0]);
    }
    else
    {
        values.push_back(value); 
    }

    tbl->SetFlags(contInfo._ddlCats[configIndex].col1, ISTable::DT_STRING |
      ISTable::CASE_INSENSE);

    tbl->Search(listOut, values, ColumnNames);
    if (listOut.empty())
    {
        return;
    }

    //try
    //{
    for (unsigned int ind = 0; ind < contInfo.\
      _ddlCats[configIndex].ddlItems.size(); ++ind)
    {
        if (tbl->IsColumnPresent(contInfo._ddlCats[configIndex].\
          ddlItems[ind].ddlItemName))
        {
            columnNames.push_back(contInfo._ddlCats[configIndex].\
              ddlItems[ind].ddlItemName);
        }
        else
        {
            // VLAD - CODE ENHANCEMENTS - FIX THIS
            columnNames.push_back(String::Empty);
        }
    }

    for (i = 0; i < (int)listOut.size(); i++)
    {
        for (unsigned int ind = 0; ind < contInfo.\
          _ddlCats[configIndex].ddlItems.size(); ++ind)
        {
            // VLAD - CODE ENHANCEMENTS - FIX THIS
            if (!columnNames[ind].empty())
            {
                //tbl->GetCell(cell, columnNames[ind], listOut[i]);
                cell = (*tbl)(listOut[i], columnNames[ind]);
            }
            combo[ind].push_back(cell);
            cell.clear();
        }
    }
    //}
    //catch (...)
    //{
      // VLAD - BUG - just continue execution. address this later
      // Reproduce in dict suite: make odb
    //}


}


DdlItemCont::DdlItemCont(FileNavigator* fnav, DicFile* dicFile,
  const string& dataBlockName, const string& itemName) :
  DdlLoadCont(fnav, dicFile, dataBlockName, itemName, RcsbDdlItem)
{
    GetItemDecendency(_decendency);
}


DdlItemCont::~DdlItemCont()
{
    _decendency.clear();
}


int DdlItemCont::Load()
{
   if (_datablock.empty() || _id.empty())
       return KEYS_NOT_SET;

    for (unsigned int i = 0; i < contInfo._ddlCats.size(); ++i)
    {
        if (contInfo._ddlCats[i].inheritance)
            LoadDdlItems(_ddlItemsStore[i], i);
        else
            DdlLoadCont::LoadDdlItems(_ddlItemsStore[i], i);
    }

    return NO_ERROR;
}


void DdlItemCont::GetItemDecendency(vector<family_member>& decendency)
/* -------------------------------------------------------------------------
   GetItemDecendency(): creates and returns a complete Family for the item
   pointed to by dataBlockName, categoryName, and itemKeyword.  The number of
   decendents in the Family is returned in the reference numD.
   ------------------------------------------------------------------------- */
{
    vector<string> ColumnNamesCh;
    vector<string> ColumnNamesPr;
    vector<string> values;
    vector<unsigned int> listOut;
    string cell;
    int num,len, num1;

    decendency.clear();

    int * parentList, i;
    int pCount,totalCount, generationNum;
    pCount = 0;
    totalCount = 0;
    generationNum = 1;
    if (_datablock.empty()) return;
  
    Block& block = _dic->GetBlock(_datablock);

    ISTable *tbl = block.GetTablePtr(DdlContInfo::CIF_DDL_CATEGORY_ITEM_LINKED);
    if(tbl == NULL) return;
    ColumnNamesCh.push_back(DdlContInfo::CIF_DDL_ITEM_CHILD_NAME);
    ColumnNamesPr.push_back(DdlContInfo::CIF_DDL_ITEM_PARENT_NAME);
    values.push_back(_id);   

    tbl->SetFlags(DdlContInfo::CIF_DDL_ITEM_CHILD_NAME, ISTable::DT_STRING |
      ISTable::CASE_INSENSE);

    tbl->Search(listOut, values, ColumnNamesCh);
    if (listOut.empty())
        return;
  
    len = 2*tbl->GetNumRows()*sizeof(int)+1;
    parentList = (int *) malloc(len);

    num1 = listOut.size();
    for (i=0;i<num1; i++) {
	 parentList[2*pCount] = listOut[i];
	 parentList[2*pCount + 1] = generationNum;
	 pCount++;
    }
    while (totalCount<pCount) {
	 values.clear();
	 cell.clear();
	 //tbl->GetCell(cell, DdlContInfo::CIF_DDL_ITEM_PARENT_NAME,
         //  parentList[2*totalCount]);
         cell = (*tbl)(parentList[2*totalCount],
           DdlContInfo::CIF_DDL_ITEM_PARENT_NAME);
	 values.push_back(cell);
	 cell.clear();
	 tbl->Search(listOut, values, ColumnNamesCh);
	 if (!listOut.empty()) {
		generationNum = parentList[2*totalCount + 1] + 1;
		num = listOut.size();
		for (i=0;i<num; i++) {
		  parentList[2*pCount] = listOut[i];
		  parentList[2*pCount + 1] = generationNum;
		  pCount++;
		}
	 }
	 totalCount++;
    }

    family_member familyMember;
    for (i = 0; i < totalCount; i++)
    {
	 //tbl->GetCell(cell, DdlContInfo::CIF_DDL_ITEM_PARENT_NAME,
         //  parentList[2*i]);
         cell = (*tbl)(parentList[2*i],
           DdlContInfo::CIF_DDL_ITEM_PARENT_NAME);
         familyMember.name = cell;
	 cell.clear();
         familyMember.generation = parentList[2 * i + 1];

         decendency.push_back(familyMember);
    }
  
    if (parentList) free(parentList);
}


void DdlItemCont::LoadDdlItems(vector<vector<string> >& combo,
  const unsigned int configIndex)
{
    if (_id.empty() || _datablock.empty())
        return;

    DdlLoadCont::LoadDdlItems(combo, configIndex);

    if (!combo[0].empty())
        return;
    else
    {
        for (unsigned int i = 0; i < _decendency.size(); i++)
        {
            DdlLoadCont::LoadDdlItems(combo, configIndex, _decendency[i].name);
            if (!combo[0].empty())
                break;
        }
        return;
    }
}


DdlDictCont::DdlDictCont(const string& dataBlockName, FileNavigator* fnav)
  : DdlLoadCont(fnav, NULL, dataBlockName, dataBlockName, RcsbDdlDict)
{
    Init();
}


DdlDictCont::DdlDictCont(const string& dataBlockName, FileNavigator* fnav,
  DicFile* dicFile) : DdlLoadCont(fnav, dicFile, dataBlockName, dataBlockName,
  RcsbDdlDict)
{
    Init();
}


DdlDictCont::~DdlDictCont()
{
    _verbose = 0;
    // VLAD - IMPLEMENT - PROPERLY IMPLEMENT THIS DESTRUCTOR
}



int DdlDictCont::Load()
{

    DdlLoadCont::Load();

    unsigned int index =
      RcsbDdlDict.GetDdlItemIndex(DdlContInfo::CIF_DDL_CATEGORY_CATEGORY,
      DdlContInfo::CIF_DDL_ITEM_ID);
    sort(_ddlItemsStore[index][0].begin(), _ddlItemsStore[index][0].end());

    index =
      RcsbDdlDict.GetDdlItemIndex(DdlContInfo::CIF_DDL_CATEGORY_SUB_CATEGORY,
      DdlContInfo::CIF_DDL_ITEM_ID);
    sort(_ddlItemsStore[index][0].begin(), _ddlItemsStore[index][0].end());

    index =
      RcsbDdlDict.GetDdlItemIndex(DdlContInfo::CIF_DDL_CATEGORY_ITEM,
      DdlContInfo::CIF_DDL_ITEM_NAME);
    sort(_ddlItemsStore[index][0].begin(), _ddlItemsStore[index][0].end());

    return NO_ERROR;

}


int DdlDictCont::Write()
{
    DdlItemCont* ilo;
    DdlLoadCont* slo;
    DdlLoadCont* clo;

    pair<unsigned int, unsigned int> indexPair =
      RcsbDdlDict.GetDdlItemIndices(DdlContInfo::CIF_DDL_CATEGORY_ITEM,
      DdlContInfo::CIF_DDL_ITEM_NAME);

    vector<uWord> itemsIndices;

    for (unsigned int i = 0; i < _ddlItemsStore[indexPair.first]\
      [indexPair.second].size(); i++)
    {
        ilo = new DdlItemCont(_fnav, _dic, _datablock,
          _ddlItemsStore[indexPair.first][indexPair.second][i]);
        ilo->Load();
        if (_verbose) ilo->Print();
        itemsIndices.push_back(ilo->Write());
        delete ilo;
    }

    indexPair = RcsbDdlDict.GetDdlItemIndices(DdlContInfo::\
      CIF_DDL_CATEGORY_SUB_CATEGORY, DdlContInfo::CIF_DDL_ITEM_ID);

    vector<uWord> subcategoriesIndices;

    for (unsigned int i = 0; i < _ddlItemsStore[indexPair.first]\
      [indexPair.second].size(); i++)
    {
        slo = new DdlLoadCont(_fnav, _dic, _datablock,
          _ddlItemsStore[indexPair.first][indexPair.second][i], RcsbDdlSubcat);
        slo->Load();
        if (_verbose) slo->Print();
        subcategoriesIndices.push_back(slo->Write());
        delete slo;
    }

    indexPair = RcsbDdlDict.GetDdlItemIndices(DdlContInfo::\
      CIF_DDL_CATEGORY_CATEGORY, DdlContInfo::CIF_DDL_ITEM_ID);

    vector<uWord> categoriesIndices;

    for (unsigned int i = 0; i < _ddlItemsStore[indexPair.first]\
      [indexPair.second].size(); i++)
    {
        clo = new DdlLoadCont(_fnav, _dic, _datablock,
          _ddlItemsStore[indexPair.first][indexPair.second][i], RcsbDdlCat);
        clo->Load();
        if (_verbose) clo->Print();
        categoriesIndices.push_back(clo->Write());
        delete clo;
    }

    _index.push_back(WriteDdlContLocations(itemsIndices));
    _index.push_back(WriteDdlContLocations(subcategoriesIndices));
    _index.push_back(WriteDdlContLocations(categoriesIndices));

    return(DdlLoadCont::Write());
}


int DdlDictCont::Read(uWord which)
{
    // HARDCODED - VLAD - FORMAT DEPENDENT
    DdlCont::Read(which, 3);

    // HARDCODED - VLAD - FORMAT DEPENDENT
    BuildDdlContainers(0, DdlContInfo::CIF_DDL_CATEGORY_ITEM,
      DdlContInfo::CIF_DDL_ITEM_NAME, _items);
    BuildDdlContainers(1, DdlContInfo::CIF_DDL_CATEGORY_SUB_CATEGORY,
      DdlContInfo::CIF_DDL_ITEM_ID, _subcategories);
    BuildDdlContainers(2, DdlContInfo::CIF_DDL_CATEGORY_CATEGORY,
      DdlContInfo::CIF_DDL_ITEM_ID, _categories);

    return NO_ERROR; 
}


DdlCont* DdlDictCont::GetDdlCont(const string& contName,
  const DdlContInfo& contInfo)
{

    if (contInfo._contInfoDescr == "item")
    {
        return(GetDdlContainers(contName, _items, contInfo));
    }
    else if (contInfo._contInfoDescr == "subcategory")
    {
        return(GetDdlContainers(contName, _subcategories, contInfo));
    }
    else if (contInfo._contInfoDescr == "category")
    {
        return(GetDdlContainers(contName, _categories, contInfo));
    }
    else
    {
        // VLAD - ERROR HANDLING
         return(NULL);
    }

    // VLAD - ERROR HANDLING
    // return(GetDdlContainers(contName, _items, contInfo));

}


int DdlDictCont::PrintDdlContainers(const string& ddlCatName,
  const string& ddlItemName, const DdlContInfo& contInfo)
{

    pair<unsigned int, unsigned int> indexPair =
      RcsbDdlDict.GetDdlItemIndices(ddlCatName, ddlItemName);

    DdlCont* ddlContP = NULL;

    for (unsigned int i = 0; i < _ddlItemsStore[indexPair.first]\
      [indexPair.second].size(); ++i)
    {
        ddlContP = GetDdlCont(_ddlItemsStore[indexPair.first]\
          [indexPair.second][i], contInfo);
        if (ddlContP != NULL)
        {
            ddlContP->Print();
            delete(ddlContP);
        }
        else
        {
            cout << "Failed to load " << contInfo._contInfoDescr <<
              " " << _ddlItemsStore[indexPair.first][indexPair.second][i] <<
              endl;
        }
    }

    return(NO_ERROR);

}


void DdlDictCont::Init()
{

}


int DdlDictCont::WriteDdlContLocations(const vector<uWord>& indices)
{
    if (indices.empty())
        return(-1);

    Word i;

    int err = _fnav->WriteUWords(indices, i);

    if (err == NO_ERROR)
    {
        // If no error, return the index that the list of idices was written to
        return(i);
    }
    else
    {
        // If error, return the error code, that is less than zero
        return(err);
    }
}


int DdlDictCont::BuildDdlContainers(unsigned int index,
  const string& ddlCatName, const string& ddlItemName,
  mapped_ptr_vector<DdlCont>& containers)
{
    if (_index[index] == 0)
        return BUILD_LIST_ERROR;

    pair<unsigned int, unsigned int> indexPair =
      RcsbDdlDict.GetDdlItemIndices(ddlCatName, ddlItemName);

    int err;

    vector<uWord> ddlContIndices;
    _fnav->GetUWords(ddlContIndices, _index[index], err);

    containers.push_back(_ddlItemsStore[indexPair.first][indexPair.second],
      ddlContIndices);
 
    return NO_ERROR;
}


void DdlDictCont::LoadDdlItems(vector<vector<string> >& combo,
  const unsigned int configIndex)
{

    ISTable* isTableP = NULL;
    vector<string> columnNames;

    if (_datablock.empty())
    {
        return;
    }

    Block& block = _dic->GetBlock(_datablock);

    isTableP = block.GetTablePtr(RcsbDdlDict._ddlCats[configIndex].ddlCatName);
    if (isTableP == NULL)
    {
        return;
    }

    for (unsigned int ind = 0; ind < RcsbDdlDict._ddlCats[configIndex].\
      ddlItems.size(); ++ind)
    {
        columnNames.push_back(RcsbDdlDict._ddlCats[configIndex].\
          ddlItems[ind].ddlItemName);
    }
  
    for (unsigned int ind = 0; ind < RcsbDdlDict._ddlCats[configIndex].\
      ddlItems.size(); ++ind)
    {
        isTableP->GetColumn(combo[ind], columnNames[ind]);
    }

}


DdlCont* DdlDictCont::GetDdlContainers(const string& contName,
  mapped_ptr_vector<DdlCont>& containers, const DdlContInfo& contInfo)
{
    unsigned int index = containers.find(contName);
    if (index == containers.size())
    {
        // VLAD_ERROR HANDLING
        return NULL;
    }

#ifdef VLAD_MODIFY
    if (containers[index] != NULL)
        return containers[index];
#else
    if (containers.is_read(contName))
        return(&(containers[index]));
#endif

#ifdef VLAD_MODIFY
    containers[index] = new DdlCont(_datablock, contName, contInfo, _fnav);
#else
    containers.set(new DdlCont(_datablock, contName, contInfo, _fnav));
#endif
    containers.read(contName);

    return(&(containers[index]));
}


DictFileCont::DictFileCont(const string& fnavFileName,
  const string& dicFileName, const string& ddlFileName, const bool verbose)
{
    IntInit(fnavFileName);

    _dicFileName = dicFileName;
    _ddlFileName = ddlFileName;
    _verbose = verbose;
}


DictFileCont::DictFileCont(const string& fnavFileName)
{
    IntInit(fnavFileName);
}


DictFileCont::~DictFileCont()
{
    _ddlFileName.clear();

    if (_ddl)
        delete (_ddl);

    if (_ddlformat)
        delete (_ddlformat);

    if (_dic)
        delete (_dic);

    if (_dicformat)
        delete (_dicformat);

    if (_fnav)
    {
      _fnav->CloseFile();
      delete _fnav;
      _fnav = NULL;
    }


    if (!_dictionaries.empty())
    {
      for (unsigned int j = 0; j < _dictionaries.size(); ++j)
        delete(&(_dictionaries[j]));
      _dictionaries.clear();
    }

}


int DictFileCont::Init()
{
    if (_ddlFileName.empty())
        return FILENAME_INVALID;

    string diags;

    _ddlformat = new ISTable("ddlformat");

    _ddl = new DdlFile(VIRTUAL_MODE, "virtual", true);

    if (_ddl == NULL)
        return ALLOCATION_ERROR;

    DDLParser* ddlParserR = NULL;

    ddlParserR = new DDLParser(_ddl, _ddlformat, _ddl->GetVerbose());

    ddlParserR->Parse(_ddlFileName, diags);

    delete(ddlParserR);

    _ddl->DataChecking(*_ddl);

    _dicformat = new ISTable("dicformat");
    _dic = new DicFile(VIRTUAL_MODE, "virtual");

    DICParser* dicParserR = NULL;

    dicParserR = new DICParser(_dic, _ddl, _dicformat, _dic->GetVerbose());

    dicParserR->Parse(_dicFileName, diags);

    delete(dicParserR);

    _dic->Compress(_ddl);

    _dic->DataChecking(*_ddl);

    vector<string> dictNames;
    _dic->GetBlockNames(dictNames);
    _dictionaries.push_back(dictNames);

    return (0);
}


int DictFileCont::Write()
{

    Word err;

    OpenFile(CREATE_MODE);

    vector<string> blockNames;

    GetDictionaryNames(blockNames);

    cout << "Number of dictionary datablocks: " << blockNames.size() << endl;

    // Write ODB file version
    // HARDCODED - VLAD - FORMAT DEPENDENT
    _fnav->WriteString(OdbFileVersion, err);
    _fnav->WriteWord(23, err);
    err = 0;

    sort(blockNames.begin(), blockNames.end());

    DdlDictCont* dlo = NULL;

    vector<uWord> list;
    Word ret = 0;
    for (unsigned int i = 0; i < blockNames.size(); i++)
    {
        dlo = new DdlDictCont(blockNames[i], _fnav, _dic);

        dlo->SetVerbose(_verbose);

        cout << "Loading dictionary object: " << blockNames[i] << endl;
        dlo->Load();

        if (_verbose != 0)
        {
            cout << "Printing dictionary object: " << blockNames[i] << endl;
            dlo->Print();
        }

        cout << "Writing dictionary object: " << blockNames[i] << endl;
        ret = dlo->Write();

        if (ret > 0)
        {
            list.push_back(ret);
        }
        else
        {
            cout << "Write dictionary object error code  = " << ret << endl;
            list.push_back(0);
        }
        delete dlo;
    }

    for (unsigned int i = 0; i < GetNumDictionaries(); i++)
    {
        cout << "Metadata used for datablock number " << i << " : " <<
          _ddl->GetFirstBlockName() << endl;
    }

    _fnav->WriteUWords(list, ret);
    _fnav->UpdateWord(ret, 1, err);

    cout << "Last index = " << ret << endl;

    _fnav->WriteStrings(blockNames, ret);

    cout << "Return code = " << 0 << endl;

    _fnav->CloseFile();

    blockNames.clear();

    list.clear();

    return(0);
}


void DictFileCont::Read()
{

    OpenFile(READ_MODE);

    int err;

    string firstString;
    // HARDCODED - VLAD - FORMAT DEPENDENT
    _fnav->GetString(firstString, 0, err);
    if (firstString != OdbFileVersion)
    {
        cout << "Cannot read old versions of dictionary object file!" << endl <<
          "Exiting program..." << endl;
        exit(1);
    }

    // HARDCODED - VLAD - FORMAT DEPENDENT
    // Location of the dictionary indices in the file.
    Word location = _fnav->GetWord(1, err);

    if (location == 0)
        // VLAD - ERROR HANDLING return BUILD_LIST_ERROR;
        return;

    vector<uWord> dictIndices;
    _fnav->GetUWords(dictIndices, location, err);

    vector<string> dictNames;
    // HARDCODED - VLAD - FORMAT DEPENDENT
    _fnav->GetStrings(dictNames, location + 1, err);

    if (dictIndices.size() != dictNames.size())
        return;
        // return BUILD_LIST_ERROR;

    _dictionaries.push_back(dictNames, dictIndices);

    //return NO_ERROR;

}


unsigned int DictFileCont::GetNumDictionaries()
{
    return(_dictionaries.size());
}


void DictFileCont::GetDictionaryNames(vector<string>& dictNames)
{
    dictNames.clear();

    if (_dic != NULL)
    {
        _dic->GetBlockNames(dictNames);
    }
    else
    {
        for (unsigned int index = 0; index < _dictionaries.size(); ++index)
        {
            dictNames.push_back(_dictionaries[index].GetName());
        }
    }
}


void DictFileCont::SetCurrentDictionary(const string& dictionaryName)
{

    unsigned int l = _dictionaries.find(dictionaryName);
    if (l == _dictionaries.size())
    // VLAD - ERROR - HANDLING - CODE
        return;

#ifdef VLAD_MODIFIED
    if (_dictionaries[l] == NULL)
    {
        _dictionaries[l] = new DdlDictCont(dictionaryName, _fnav);
        _dictionaries.read(dictionaryName);
    }
#endif

    if (!_dictionaries.is_read(dictionaryName))
    {
        _dictionaries.set(new DdlDictCont(dictionaryName, _fnav));
        _dictionaries.read(dictionaryName);
    }

    _currDictContP = &(_dictionaries[l]);

}


DdlDictCont* DictFileCont::GetCurrDictCont()
{
    return(_currDictContP);
}


void DictFileCont::Print()
{
    for (uWord i = 0; i < _dictionaries.size(); i++)
    {
        SetCurrentDictionary(_dictionaries[i].GetName());

        _currDictContP->PrintDdlContainers(DdlContInfo::\
          CIF_DDL_CATEGORY_CATEGORY, DdlContInfo::CIF_DDL_ITEM_ID, RcsbDdlCat);
        _currDictContP->PrintDdlContainers(DdlContInfo::\
          CIF_DDL_CATEGORY_SUB_CATEGORY, DdlContInfo::CIF_DDL_ITEM_ID,
          RcsbDdlSubcat);
        _currDictContP->PrintDdlContainers(DdlContInfo::CIF_DDL_CATEGORY_ITEM,
          DdlContInfo::CIF_DDL_ITEM_NAME, RcsbDdlItem);

        _currDictContP->Print();    
    }
}


void DictFileCont::IntInit(const string& fnavFileName)
{
    _verbose = false;

    _ddl = NULL;
    _dic = NULL;
    _fnav = NULL;

    _ddlformat = NULL;
    _dicformat = NULL;

    _dictionaries.clear();

    _currDictContP = NULL;

    _fnavFileName = fnavFileName;

    _fnav = new FileNavigator;
}


void DictFileCont::OpenFile(const eFileMode fileMode)
{
    _fnav->OpenFile(_fnavFileName, fileMode);

    _fnav->ReadFileHeader();
}

